SystemOrganization addCategory: #SplitJoin!

!SequenceableCollection methodsFor: '*splitjoin' stamp: 'onierstrasz 4/12/2009 19:58'!
appendTo: aCollection
	"double dispatch for join:"
	^ aCollection addAllLast: self! !

!SequenceableCollection methodsFor: '*splitjoin' stamp: 'onierstrasz 4/12/2009 20:19'!
join: aCollection
	"NB: this implementation only works for Array, since WriteStreams only work for Arrays and Strings. (!!)
	Overridden in OrderedCollection and SortedCollection."
	^ self class
		streamContents: [:stream | aCollection
				do: [:each | each joinTo: stream]
				separatedBy: [stream nextPutAll: self]]! !

!SequenceableCollection methodsFor: '*splitjoin' stamp: 'onierstrasz 4/12/2009 19:58'!
joinTo: stream
	"double dispatch for join:"
	^ stream nextPutAll: self! !

!SequenceableCollection methodsFor: '*splitjoin' stamp: 'onierstrasz 4/10/2009 22:53'!
joinUsing: joiner
	"joiner - character, string or sequenceable collection
	returns collection of the same collection class as 'joiner', or a String"
	^ joiner join: self! !

!SequenceableCollection methodsFor: '*splitjoin' stamp: 'onierstrasz 4/10/2009 22:54'!
joinUsing: joiner last: last 
	"#(1 2 3 4) joinUsing: ', ' last: 'and'. => '1, 2, 3 and 4"
	^ last join: (Array
				with: (joiner join: self allButLast)
				with: self last)! !

!SequenceableCollection methodsFor: '*splitjoin' stamp: 'onierstrasz 4/12/2009 20:57'!
split: aSequencableCollection
	| result position oldPosition |
	result := OrderedCollection new.
	position := 1.
	oldPosition := position.
	position := aSequencableCollection indexOfSubCollection: self startingAt: position.
	[ position > 0 ] whileTrue: [
		result add: (aSequencableCollection copyFrom: oldPosition to: position - 1).
		position := position + self size.
		oldPosition := position.
		position := aSequencableCollection indexOfSubCollection: self startingAt: position.
	].
	result add: (aSequencableCollection copyFrom: oldPosition to: aSequencableCollection size).
	^ result

! !

!SequenceableCollection methodsFor: '*splitjoin' stamp: 'onierstrasz 4/12/2009 20:20'!
splitOn: splitter 
	"splitter - can be a subsequence, a Block or a Regex (String receiver only).
	Any other object used as a splitter is treated as an Array containing that object."
	^ splitter split: self! !

!Character methodsFor: '*splitjoin' stamp: 'onierstrasz 4/10/2009 22:51'!
join: aSequenceableCollection
	^ self asString join: aSequenceableCollection
! !

!OrderedCollection methodsFor: '*splitjoin' stamp: 'onierstrasz 4/12/2009 19:44'!
join: aCollection 
	| result |
	result := self class new.
	aCollection
		do: [:each | each appendTo: result]
		separatedBy: [self appendTo: result].
	^ result! !

!String methodsFor: '*splitjoin' stamp: 'onierstrasz 4/12/2009 20:21'!
join: aCollection 
	^ String
		streamContents: [:stream | aCollection
				do: [:each | stream nextPutAll: each asString] "NB: coerce elements to Strings"
				separatedBy: [stream nextPutAll: self]]! !

!SortedCollection methodsFor: '*splitjoin' stamp: 'onierstrasz 4/12/2009 19:57'!
join: aCollection 
	"Curiously addAllLast: does not trigger a reSort, so we must do it here."
	^ (super join: aCollection) reSort; yourself
! !

TestCase subclass: #SplitJoinTest
	instanceVariableNames: 'eg stringToSplit'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SplitJoin'!

!SplitJoinTest commentStamp: 'onierstrasz 4/12/2009 20:33' prior: 0!
"Evaluate me to view the SplitJoin documentation:"

self showDocumentation

"CHANGES LOG:
- merged implementations by Keith Hodges (Join) and Damiena Pollet
  and Oscar Nierstrasz (RubyShards) into SplitJoin package
- moved all extension methods to *splitjoin method category
- merged all tests into SplitJoinTest
- fixed protocol in SequenceableCollection to splitOn: and joinUsing:
  and split: join: for splitters and joiners
- added Object>>joinTo: aStream and SequenceableCollection>>joinTo: aStream
  to support joining of either sequences or sequences of sequences
- prepared some documentation
- added systematic tests for all split/join cases
- added Object>>join:
- prepared split/join tests for all 16 cases
- prepares split+join tests for 4 standard cases
- reviewed/merged old tests
- changed splitjoin tests to use different joiner
- added separate test for split+join=id cases
- adapted documentation -- join result type is type of joiner or array or string
- fix split tests to check result asOrderedCollection
- added split tests for OrderedCollection and SortedCollection
- new join: method for OrderedCollection and SortedCollection
  (uses appendTo: in Object and SequenceableCollection)
- reviewed all split: implementations -- removed unnecessary helper methods
- check boundary conditions -- split on empty sequence ...
"!

!SplitJoinTest class methodsFor: 'documentation' stamp: 'onierstrasz 4/11/2009 14:19'!
documentation
	<ignoreForCoverage>
	"self showDocumentation"
	
	^ '"This package provides functionality for splitting and joining strings similarly to that offered by Perl, Python and Ruby. In addition, split and join work for any kind of sequence, and offer the possibility to split strings using regular expressions.

Suppose object A is a SequenceableCollection of elements of type T (let''s call its type S[T]), and B is a splitter of type either T or S[T]. Then A splitOn: B yields an object C of type OC[S[T]] (OC=OrderedCollection). B can be either an element or a subsequence, so:"

	''abracadabra'' splitOn: $b.
	"-> an OrderedCollection(''a'' ''racada'' ''ra'')"

	''abracadabra'' splitOn: ''ca''.
	"-> an OrderedCollection(''abra'' ''dabra'')"

"A splitOn: B is the same as B split: A, so:"

	''ab'' split: ''abracadabra''.
	"-> an OrderedCollection('''' ''racad'' ''ra'')"

"joinUsing: and join: are the inverse of splitOn: and split:. You may use either of the forms:
	C joinUsing: B
or:
	B join: C
The result is normally the same type as the joiner. If the joiner (B) is of type S[T] and C is of type OC[S[T]] or Array[S[T]], the result is also of type S[T]. For example:"

	'' loves '' join: #(''john'' ''jane'' ''jack'').
	"-> ''john loves jane loves jack''"

"An object that is not a sequence can also be used as a joiner, in which case the final result is an array. Also the collection being joined may contain elements that are not sequences, in which case they are treated as singleton arrays."

	1 join: ''hello''.
	"-> #($h 1 $e 1 $l 1 $l 1 $o)"

"Note that in the case of characters and strings being used as joiners, the result is always a string."

"In general:
	((A splitOn: B) joinUsing: B) = A
or:
	(B join: (B split: A)) = A
so:"

	(''r'' join: (''r'' split: ''abracadabra'')) = ''abracadabra''.
	"->  true"

"Although the main application is for Strings, split and join will work for any kinds of SequenceableCollections, including Arrays and Intervals."

	[:n | n isPrime] split: (2 to: 20).
	"-> an OrderedCollection(#() #() #(4) #(6) #(8 9 10) #(12) #(14 15 16) #(18) #(20))"

"In addition, the following special cases are supported:

- split an S[T] using a block of type T->Boolean:"

	''abracadabra'' splitOn: [:x | ''bc'' includes: x].
	"-> an OrderedCollection(''a'' ''ra'' ''ada'' ''ra'')"

"- split a String using a regex:"

	''abracadabra'' splitOn: ''[bcdr]+'' asRegex.
	"-> an OrderedCollection(''a'' ''a'' ''a'' ''a'' ''a'')"

"- join all elements of an S[T] yielding another S[T]: '' ''- "

	''abracadabra'' joinUsing: '':''.
	" -> ''a:b:r:a:c:a:d:a:b:r:a''"

"Note that in these cases the rule that (B join: (B split: A)) = A does not hold, either because B maps to different elements, or because there is no A to start with."

	$: join: (1 to: 4).
	"-> ''1:2:3:4''"

"Finally, there is the convenient utility method joinUsing:last:, used as follows:"

	(1 to: 5) joinUsing: '', '' last: '' and ''.
	"-> ''1, 2, 3, 4 and 5''"

"This package merges and generalizes functionality from the Join package by Keith Hodges and the RubyShards package by Damien Pollet and Oscar Nierstrasz. "
'! !

!SplitJoinTest class methodsFor: 'sunitgui' stamp: 'onierstrasz 4/12/2009 18:50'!
packageNamesUnderTest
	<ignoreForCoverage>
	^ #('SplitJoin')! !

!SplitJoinTest class methodsFor: 'documentation' stamp: 'onierstrasz 4/10/2009 23:41'!
showDocumentation
	<ignoreForCoverage>
	SHWorkspace new contents: self documentation; openLabel: 'SplitJoin Documentation'.! !

!SplitJoinTest methodsFor: 'testing' stamp: 'onierstrasz 4/10/2009 21:09'!
assert: result equals: expected 
	result = expected
		ifFalse: [self signalFailure: 'Assertion failed: (' , result asString , ') ~= (' , expected asString , ')']! !

!SplitJoinTest methodsFor: 'running' stamp: 'onierstrasz 4/12/2009 19:22'!
testJoinArrayUsingArray
	self assert: ((1 to: 4) joinUsing: #(8 9))
		equals: #(1 8 9 2 8 9 3 8 9 4)! !

!SplitJoinTest methodsFor: 'running' stamp: 'onierstrasz 4/10/2009 23:14'!
testJoinArrayUsingChar
	self assert: ((1 to: 4) joinUsing: $:)
		equals:  '1:2:3:4'! !

!SplitJoinTest methodsFor: 'running' stamp: 'onierstrasz 4/11/2009 00:11'!
testJoinArrayUsingObject
	self assert: ((1 to: 4) joinUsing: 0)
		equals: #(1 0 2 0 3 0 4)! !

!SplitJoinTest methodsFor: 'running' stamp: 'onierstrasz 4/12/2009 19:23'!
testJoinArrayUsingOrderedCollection
	self assert: ((1 to: 4) joinUsing: #(8 9) asOrderedCollection)
		equals: #(1 8 9 2 8 9 3 8 9 4) asOrderedCollection! !

!SplitJoinTest methodsFor: 'running' stamp: 'onierstrasz 4/12/2009 19:46'!
testJoinArrayUsingSortedCollection
	self assert: ((1 to: 4) joinUsing: #(8 9) asSortedCollection)
		equals: #(1 8 9 2 8 9 3 8 9 4) asSortedCollection! !

!SplitJoinTest methodsFor: 'running' stamp: 'onierstrasz 4/10/2009 23:14'!
testJoinArrayUsingString
	self assert: ((1 to: 4) joinUsing: '--')
		equals: '1--2--3--4'! !

!SplitJoinTest methodsFor: 'running' stamp: 'onierstrasz 4/10/2009 23:15'!
testJoinStringUsingChar
	self assert: ('splitjoin' joinUsing: $-)
		equals: 's-p-l-i-t-j-o-i-n'! !

!SplitJoinTest methodsFor: 'running' stamp: 'onierstrasz 4/10/2009 23:17'!
testJoinStringUsingString
	self assert: ('bda' joinUsing: 'an')
		equals: 'bandana'! !

!SplitJoinTest methodsFor: 'running' stamp: 'onierstrasz 4/10/2009 23:17'!
testJoinUsingLastOnArray
	self assert: ((1 to: 4) joinUsing: ', ' last: ' and ')
		equals: '1, 2, 3 and 4'! !

!SplitJoinTest methodsFor: 'running' stamp: 'onierstrasz 4/10/2009 22:38'!
testJoinUsingLastOnArrayOfStrings
	self
		assert: (#('Squeak is modern' 'open source' 'highly portable' 'fast' 'full-featured' ) joinUsing: ', ' last: ' and ')
		equals: 'Squeak is modern, open source, highly portable, fast and full-featured'! !

!SplitJoinTest methodsFor: 'running' stamp: 'onierstrasz 4/12/2009 19:24'!
testSplitArrayOnBlock
	self assert: ((1 to: 10) asArray splitOn: [:n| n even])
		equals: #(#(1) #(3) #(5) #(7) #(9) #()) asOrderedCollection! !

!SplitJoinTest methodsFor: 'running' stamp: 'onierstrasz 4/12/2009 19:24'!
testSplitArrayOnElement
	self assert: ((1 to: 10) asArray splitOn: 4)
		equals: #(#(1 2 3) #(5 6 7 8 9 10)) asOrderedCollection
! !

!SplitJoinTest methodsFor: 'running' stamp: 'onierstrasz 4/12/2009 19:25'!
testSplitArrayOnSequence
	self assert: ((1 to: 10) asArray splitOn: (4 to: 5))
		equals: #(#(1 2 3) #(6 7 8 9 10)) asOrderedCollection
! !

!SplitJoinTest methodsFor: 'running' stamp: 'onierstrasz 4/12/2009 20:32'!
testSplitJoinBoundaryCases
	"Empty splitter, joiner or sequence."
	self assert: (#() join: (#() split: #())) equals: #().
	self assert: ('' join: ('' split: '')) equals: ''.
	self assert: ('' join: ('.' asRegex split: '')) equals: ''. "NB: Doesn't work with empty regex"

	"Overlapping splitters, or at end of sequence"
	self assert: ('an' join: ('an' split: 'banana')) equals: 'banana'.
	self assert: ('na' join: ('na' split: 'banana')) equals: 'banana'.
	self assert: ('ana' join: ('ana' split: 'banana')) equals: 'banana'.
! !

!SplitJoinTest methodsFor: 'running' stamp: 'onierstrasz 4/12/2009 20:07'!
testSplitJoinIdentity
	| array string |
	array := #(5 1 4 1 3 1 2 1).
	string := 'how now brown cow'.
	self assert: (1 join: (1 split: array)) equals: array.
	self assert: (#(1 3) join: (#(1 3) split: array)) equals: array.
	self assert: ($o join: ($o split: string)) equals: string.
	self assert: ('ow' join: ('ow' split: string)) equals: string.! !

!SplitJoinTest methodsFor: 'running' stamp: 'onierstrasz 4/11/2009 10:31'!
testSplitJoinOnElement
	self assert: (0 join: (3 split: #(1 2 3 4 5)))
		equals: #(1 2 0 4 5)! !

!SplitJoinTest methodsFor: 'running' stamp: 'onierstrasz 4/11/2009 10:33'!
testSplitJoinOnSequence
	self assert: (#(6 6 6) join: (#(3 4) split: #(1 2 3 4 5)))
		equals: #(1 2 6 6 6 5)! !

!SplitJoinTest methodsFor: 'running' stamp: 'onierstrasz 4/11/2009 10:32'!
testSplitJoinStringOnChar
	self assert: ($: join: (Character space split: 'how now brown cow?'))
		equals: 'how:now:brown:cow?'! !

!SplitJoinTest methodsFor: 'running' stamp: 'onierstrasz 4/11/2009 10:36'!
testSplitJoinStringOnString
	self assert: ('oo' join: ('ow' split: 'how now brown cow?'))
		equals: 'hoo noo broon coo?'! !

!SplitJoinTest methodsFor: 'running' stamp: 'onierstrasz 4/12/2009 19:26'!
testSplitOrderedCollectionOnElement
	self assert: (((1 to: 10) asOrderedCollection) splitOn: 4)
		equals: {(1 to: 3) asOrderedCollection . (5 to: 10) asOrderedCollection} asOrderedCollection.
! !

!SplitJoinTest methodsFor: 'running' stamp: 'onierstrasz 4/12/2009 19:26'!
testSplitOrderedCollectionOnOrderedCollection
	self assert: (((1 to: 10) asOrderedCollection) splitOn: ((4 to: 5) asOrderedCollection))
		equals: {(1 to: 3) asOrderedCollection . (6 to: 10) asOrderedCollection} asOrderedCollection.
! !

!SplitJoinTest methodsFor: 'running' stamp: 'onierstrasz 4/12/2009 19:27'!
testSplitSortedCollectionOnElement
	self assert: (((1 to: 10) asSortedCollection) splitOn: 4)
		equals: {(1 to: 3) asSortedCollection . (5 to: 10) asSortedCollection} asOrderedCollection.
! !

!SplitJoinTest methodsFor: 'running' stamp: 'onierstrasz 4/12/2009 19:26'!
testSplitSortedCollectionOnSortedCollection
	self assert: (((1 to: 10) asSortedCollection) splitOn: ((4 to: 5) asSortedCollection))
		equals: {(1 to: 3) asSortedCollection . (6 to: 10) asSortedCollection} asOrderedCollection.

! !

!SplitJoinTest methodsFor: 'running' stamp: 'onierstrasz 4/12/2009 18:52'!
testSplitStringOnBlock
	self assert: ('foobar' splitOn: [:ch | 'aeiou' includes: ch])
		equals: #('f' '' 'b' 'r') asOrderedCollection! !

!SplitJoinTest methodsFor: 'running' stamp: 'onierstrasz 4/12/2009 18:53'!
testSplitStringOnChar
	self assert: ('does eat oats and lambs eat oats' splitOn: Character space)
		equals: #('does' 'eat' 'oats' 'and' 'lambs' 'eat' 'oats') asOrderedCollection! !

!SplitJoinTest methodsFor: 'running' stamp: 'onierstrasz 4/12/2009 18:53'!
testSplitStringOnRegex
	self assert: ('foobar' splitOn: '[aeiou]+' asRegex)
		equals: #('f' 'b' 'r') asOrderedCollection! !

!SplitJoinTest methodsFor: 'running' stamp: 'onierstrasz 4/12/2009 18:53'!
testSplitStringOnSubstring
	self assert: ('the banana man can can bananas' splitOn: 'an')
		equals: #('the b' '' 'a m' ' c' ' c' ' b' '' 'as') asOrderedCollection! !

!BlockContext methodsFor: '*splitjoin' stamp: 'onierstrasz 4/12/2009 20:12'!
split: aSequenceableCollection 
	| result position |
	result := OrderedCollection new.
	position := 1.
	aSequenceableCollection
		withIndexDo: [:element :idx |
			(self value: element)
				ifTrue: [result add: (aSequenceableCollection copyFrom: position to: idx - 1).
					position := idx + 1]].
	result add: (aSequenceableCollection copyFrom: position to: aSequenceableCollection size).
	^ result! !

!RxMatcher methodsFor: '*splitjoin' stamp: 'onierstrasz 4/12/2009 20:16'!
split: aString
	| result lastPosition |
	result := OrderedCollection new.
	stream := aString readStream.
	lastPosition := stream position.
	[ self searchStream: stream ] whileTrue:
		[ result add: (aString copyFrom: lastPosition+1 to: (self subBeginning: 1)).
		self assert: lastPosition < stream position description: 'Regex cannot match null string'.
		lastPosition := stream position ].
	result add: (aString copyFrom: lastPosition+1 to: aString size).
	^ result! !

!Object methodsFor: '*splitjoin' stamp: 'onierstrasz 4/12/2009 19:58'!
appendTo: aCollection
	"double dispatch for join:"
	^ aCollection addLast: self! !

!Object methodsFor: '*splitjoin' stamp: 'onierstrasz 4/10/2009 22:50'!
join: aSequenceableCollection
	^ (Array with: self) join: aSequenceableCollection! !

!Object methodsFor: '*splitjoin' stamp: 'onierstrasz 4/12/2009 19:58'!
joinTo: stream
	"double dispatch for join:"
	^ stream nextPut: self! !

!Object methodsFor: '*splitjoin' stamp: 'onierstrasz 4/10/2009 22:49'!
split: aSequenceableCollection
	^ (Array with: self) split: aSequenceableCollection! !
